from datetime import timedelta
from unittest.mock import patch

from django.core.files.uploadedfile import SimpleUploadedFile


def test_profile_detail(client, db, profile, user):
    response = client.get(f"/profiles/{profile.pk}/")
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == user.username


def test_profile_delete(client, db, profile, user):
    response = client.delete(f"/profiles/{profile.pk}/")
    assert response.status_code == 204, response.content
    assert response.content == b""


def test_profile_list(client, db, profile, user):
    response = client.get("/profiles/")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


def test_profile_list_filter(client, db, profile, user):
    response = client.get(f"/profiles/?name={user.first_name} {user.last_name}")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


def test_profile_list_icontains_filter(client, db, profile, user):
    response = client.get(f"/profiles/?name__icontains={user.first_name}")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


def test_profile_list_annotation_filter(client, db, profile_factory):
    profile_factory.create(user__date_joined="2020-01-01T00:00:00Z")
    profile_factory.create(user__date_joined="2020-12-01T00:00:00Z")
    response = client.get("/profiles/?dateJoined__gte=2020-06-01T00:00:00Z")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1


def test_profile_list_and_filter(client, db, profile_factory, tag_factory):
    tag1, tag2, tag3 = tag_factory.create_batch(3)
    profile_factory.create(tags=[tag1])
    profile_factory.create(tags=[tag2])
    profile_factory.create(tags=[tag1, tag2])
    profile_factory.create(tags=[tag3])
    profile_factory.create()
    response = client.get(f"/profiles/?tags={tag1.pk}&tags={tag2.pk}")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1


def test_profile_list_in_array_filter(client, db, profile, user):
    response = client.get(f"/profiles/?name__in={user.first_name} {user.last_name}&name__in=Din Djarin")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


def test_profile_list_in_string_filter(client, db, profile, user):
    response = client.get(f"/profiles/?name__in={user.first_name} {user.last_name},Din Djarin")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 1
    assert result["profiles"][0]["username"] == user.username


def test_profile_list_or_filter(client, db, profile_factory, tag_factory):
    tag1, tag2, tag3 = tag_factory.create_batch(3)
    profile_factory.create(tags=[tag1])
    profile_factory.create(tags=[tag2])
    profile_factory.create(tags=[tag1, tag2])
    profile_factory.create(tags=[tag3])
    profile_factory.create()
    response = client.get(f"/profiles/?tags__in={tag1.pk}&tags__in={tag2.pk}")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 3


def test_profile_list_negated_filter(client, db, profile, user):
    response = client.get(f"/profiles/?name!={user.first_name} {user.last_name}")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 0


def test_profile_list_negated__icontains__filter(client, db, profile, user):
    response = client.get(f"/profiles/?name__icontains!={user.first_name}")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 0


def test_profile_list_not_in_array_filter(client, db, profile, user):
    response = client.get(f"/profiles/?name__in!={user.first_name} {user.last_name}&name__in!=Din Djarin")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 0


def test_profile_list_not_in_string_filter(client, db, profile, user):
    response = client.get(f"/profiles/?name__in!={user.first_name} {user.last_name},Din Djarin")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 0


def test_profile_list_subset_filter(client, db, profile, user):
    response = client.get(f"/profiles/subset/?name={user.first_name} {user.last_name}")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["profiles"]) == 0


@patch('django.core.files.storage.FileSystemStorage.save')
def test_profile_multipart_create(mock_save, client, db, role, user):
    avatar = SimpleUploadedFile("avatar.jpg", b"", content_type="image/jpeg")
    mock_save.return_value = "avatar.jpg"
    payload = dict(avatar=avatar, role=role.pk, user=user.pk)
    response = client.post(f"/profiles/", payload)
    result = response.json()
    assert response.status_code == 201, result
    assert result["avatar"] == "/avatar.jpg"
    assert result["role"]["id"] == role.pk
    assert result["role"]["name"] == role.name
    assert result["user"]["username"] == user.username


@patch('django.core.files.storage.FileSystemStorage.save')
def test_profile_multipart_patch(mock_save, client, db, profile, role, user):
    avatar = SimpleUploadedFile("avatar.jpg", b"", content_type="image/jpeg")
    mock_save.return_value = "avatar.jpg"
    payload = dict(avatar=avatar, role=role.pk, user=user.pk)
    response = client.patch(f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert result["avatar"] == "/avatar.jpg"
    assert result["role"]["id"] == role.pk
    assert result["role"]["name"] == role.name
    assert result["user"]["username"] == user.username


@patch('django.core.files.storage.FileSystemStorage.save')
def test_profile_multipart_put(mock_save, client, db, profile, role, user):
    avatar = SimpleUploadedFile("avatar.jpg", b"", content_type="image/jpeg")
    mock_save.return_value = "avatar.jpg"
    payload = dict(avatar=avatar, role=role.pk, user=user.pk)
    response = client.put(f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert result["avatar"] == "/avatar.jpg"
    assert result["role"]["id"] == role.pk
    assert result["role"]["name"] == role.name
    assert result["user"]["username"] == user.username


def test_profile_patch_fk(client, db, profile, role, team):
    payload = dict(role=role.pk, team=team.pk)
    response = client.patch(f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert result["role"]["name"] == role.name
    assert result["team"]["name"] == team.name


def test_profile_patch_fk_invalid_role(client, db, profile, role, team):
    response = client.patch(f"/profiles/{profile.pk}/", dict(role=123))
    result = response.json()
    assert response.status_code == 422, result
    assert result["message"] == "Invalid role"


def test_profile_patch_fk_role_is_not_nullable(client, db, profile, role, team):
    response = client.patch(f"/profiles/{profile.pk}/", dict(role=None))
    result = response.json()
    assert response.status_code == 422, result
    assert result["message"] == "Invalid role"


def test_profile_patch_fk_team_is_nullable(client, db, profile, role, team):
    response = client.patch(f"/profiles/{profile.pk}/", dict(team=None))
    result = response.json()
    assert response.status_code == 200, result
    assert result["team"] is None


def test_profile_patch_m2m(client, db, profile, tag):
    response = client.patch(f"/profiles/{profile.pk}/", dict(tags=[tag.pk]))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["tags"]) == 1
    assert result["tags"][0]["id"] == tag.pk
    assert result["tags"][0]["name"] == tag.name


def test_profile_patch_m2m_can_be_empty(client, db, profile, tag):
    response = client.patch(f"/profiles/{profile.pk}/", dict(tags=[]))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["tags"]) == 0


def test_profile_patch_m2m_is_not_nullable(client, db, profile, tag):
    response = client.patch(f"/profiles/{profile.pk}/", dict(tags=None))
    result = response.json()
    assert response.status_code == 422, result
    assert "tags accepts an array, got <class 'NoneType'> None" in result["message"]


def test_profile_patch_m2m_must_be_pks(client, db, profile, tag):
    payload = dict(tags=["invalid"])
    response = client.patch(f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid tags" in result["message"]


def test_profile_patch_m2m_through(client, db, profile, skill):
    payload = dict(skills=[dict(id=skill.pk, rating=4)])
    response = client.patch(f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["skills"]) == 1
    assert result["skills"][0]["id"] == skill.pk
    assert result["skills"][0]["name"] == skill.name
    assert result["skills"][0]["rating"] == 4


def test_profile_patch_m2m_through_can_be_empty(client, db, profile, skill):
    response = client.patch(f"/profiles/{profile.pk}/", dict(skills=[]))
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["skills"]) == 0


def test_profile_patch_m2m_through_is_not_nullable(client, db, profile, skill):
    response = client.patch(f"/profiles/{profile.pk}/", dict(skills=None))
    result = response.json()
    assert response.status_code == 422, result
    assert "skills accepts an array, got <class 'NoneType'> None" in result["message"]


def test_profile_patch_m2m_through_must_be_dicts(client, db, profile, skill):
    payload = dict(skills=["invalid"])
    response = client.patch(f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid skills" == result["message"]


def test_profile_patch_m2m_through_ids_must_be_pks(client, db, profile, skill):
    payload = dict(skills=[dict(id="invalid")])
    response = client.patch(f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid skills" in result["message"]


def test_profile_patch_m2m_through_required_fields(client, db, profile, skill):
    payload = dict(skills=[dict(id=skill.pk)])
    response = client.patch(f"/profiles/{profile.pk}/", payload)
    result = response.json()
    assert response.status_code == 422, result
    assert "Invalid skills" in result["message"]


def test_user_detail(client, db, user):
    response = client.get(f"/users/{user.pk}/")
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == user.username


def test_user_list(client, db, user):
    response = client.get("/users/")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 1
    assert result["users"][0]["username"] == user.username


def test_user_list_filters(client, db, user_factory):
    january = "2021-01-01T00:00:00Z"
    february = "2021-02-01T00:00:00Z"
    march = "2021-03-01T00:00:00Z"

    user1 = user_factory.create(date_joined=january)
    user2 = user_factory.create(date_joined=february)
    user3 = user_factory.create(email="test3@test.com", date_joined=march)

    # filter by email
    response = client.get(f"/users/?email={user1.email}")
    assert len(response.json()["users"]) == 1

    # filter by string
    response = client.get("/users/?q=example.com")
    assert len(response.json()["users"]) == 2
    response = client.get(f"/users/?q={user3.email}")
    assert len(response.json()["users"]) == 1

    # filter by username -- filter ignored, username is not in filter_fields
    response = client.get(f"/users/?username={user2.username}")
    assert len(response.json()["users"]) == 3

    # filter by date joined
    gte, lte = "2021-02-01T00:00:00Z", "2021-02-15T00:00:00Z"
    response = client.get(f"/users/?dateJoined__gte={gte}&date_joined__lte={lte}")
    assert len(response.json()["users"]) == 1
    assert response.json()["users"][0]["username"] == user2.username


def test_user_list_sort_asc(client, db, user_factory):
    user_factory.create(username="a")
    user_factory.create(username="b")
    response = client.get("/users/?sort=id")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 2
    assert result["users"][0]["username"] == "a"
    assert result["users"][1]["username"] == "b"


def test_user_list_sort_desc(client, db, user_factory):
    user_factory.create(username="a")
    user_factory.create(username="b")
    response = client.get("/users/?sort=-id")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 2
    assert result["users"][0]["username"] == "b"
    assert result["users"][1]["username"] == "a"


def test_user_list_multisort(client, now, db, user_factory):
    user_factory.create(username="a", date_joined=now)
    user_factory.create(username="b", date_joined=now - timedelta(hours=1))
    user_factory.create(username="c", date_joined=now)
    user_factory.create(username="d", date_joined=now)
    response = client.get("/users/?sort=dateJoined&sort=-id")
    result = response.json()
    assert response.status_code == 200, result
    assert len(result["users"]) == 4
    assert result["users"][0]["username"] == "b"
    assert result["users"][1]["username"] == "d"
    assert result["users"][2]["username"] == "c"
    assert result["users"][3]["username"] == "a"


def test_user_patch(client, db, user):
    payload = dict(username="testtest", email="something@example.com")
    response = client.patch(f"/users/{user.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == "testtest"
    assert result["email"] == "something@example.com"


def test_user_update(client, db, user):
    payload = dict(username="testtest", email="something@example.com")
    response = client.put(f"/users/{user.pk}/", payload)
    result = response.json()
    assert response.status_code == 200, result
    assert result["username"] == "testtest"
    assert result["email"] == "something@example.com"



from worf.serializers import fields, Serializer


class UserSerializer(Serializer):
    last_login = fields.DateTime(dump_only=True)
    date_joined = fields.DateTime(dump_only=True)

    class Meta:
        fields = [
            "username",
            "last_login",
            "date_joined",
            "email",
        ]


class ProfileSerializer(Serializer):
    username = fields.Function(lambda obj: obj.user.username)
    email = fields.Function(lambda obj: obj.user.email)
    role = fields.Nested("RoleSerializer")
    skills = fields.Nested("RatedSkillSerializer", attribute="ratedskill_set", many=True)
    team = fields.Nested("TeamSerializer")
    tags = fields.Nested("TagSerializer", many=True)
    user = fields.Nested("UserSerializer")

    class Meta:
        fields = [
            "username",
            "avatar",
            "email",
            "phone",
            "role",
            "skills",
            "team",
            "tags",
            "user",
        ]


class RatedSkillSerializer(Serializer):
    id = fields.Function(lambda obj: obj.skill.id)
    name = fields.Function(lambda obj: obj.skill.name)

    class Meta:
        fields = ["id", "name", "rating"]


class RoleSerializer(Serializer):
    class Meta:
        fields = ["id", "name"]


class TagSerializer(Serializer):
    class Meta:
        fields = ["id", "name"]


class TeamSerializer(Serializer):
    class Meta:
        fields = ["id", "name"]



from django.contrib.auth.models import User

import factory
from factory.django import DjangoModelFactory

from tests.models import Profile, Role, Skill, Tag, Team


class ProfileFactory(DjangoModelFactory):
    user = factory.SubFactory("tests.factories.UserFactory")
    role = factory.SubFactory("tests.factories.RoleFactory")
    team = None

    @factory.post_generation
    def tags(self, create, extracted, **kwargs):
        if create and extracted:
            self.tags.set(extracted)

    class Meta:
        model = Profile


class RoleFactory(DjangoModelFactory):
    name = factory.Sequence(lambda i: f"Role {i}")

    class Meta:
        model = Role


class SkillFactory(DjangoModelFactory):
    name = factory.Sequence(lambda i: f"Skill {i}")

    class Meta:
        model = Skill


class TagFactory(DjangoModelFactory):
    name = factory.Sequence(lambda i: f"Tag {i}")

    class Meta:
        model = Tag


class TeamFactory(DjangoModelFactory):
    name = factory.Sequence(lambda i: f"Team {i}")

    class Meta:
        model = Team


class UserFactory(DjangoModelFactory):
    username = factory.Sequence(lambda i: f"user-{i}")
    email = factory.Sequence(lambda i: f"user-{i}@example.com")
    first_name = "Boba"
    last_name = factory.Sequence(lambda i: f"Fett {i}")

    class Meta:
        model = User



import subprocess

from worf import __version__


def get_current_version():
    try:
        hash = (
            subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
            .strip()
            .decode()
        )
        return f"{__version__}@{hash}"
    except:  # pragma: no cover # noqa E722 Dont crash for any reason whatsoever
        return __version__



import pytest
from pytest_factoryboy import register


def pytest_configure():
    """Initialize Django settings."""
    import django
    from django.conf import settings

    settings.configure(
        SECRET_KEY="secret",
        DEBUG=True,
        INSTALLED_APPS=[
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
            "django.contrib.staticfiles",
            "tests",
            "worf",
        ],
        MIDDLEWARE=[
            "django.middleware.common.CommonMiddleware",
            "django.contrib.sessions.middleware.SessionMiddleware",
            "django.contrib.auth.middleware.AuthenticationMiddleware",
            "django.contrib.messages.middleware.MessageMiddleware",
        ],
        ROOT_URLCONF="tests.urls",
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": "db.sqlite3",
            }
        },
        PASSWORD_HASHERS=["django.contrib.auth.hashers.MD5PasswordHasher"],
        TEMPLATES = [
            {
                "BACKEND": "django.template.backends.django.DjangoTemplates",
                "APP_DIRS": True,
            },
        ],
        TIME_ZONE="UTC",
        USE_I18N=True,
        USE_L10N=True,
        USE_TZ=True,
        STATIC_URL="/static/",
        WORF_API_NAME="Test API"
    )

    django.setup()

    from tests.factories import (
        ProfileFactory,
        RoleFactory,
        SkillFactory,
        TagFactory,
        TeamFactory,
        UserFactory,
    )

    register(ProfileFactory, "profile")
    register(RoleFactory, "role")
    register(SkillFactory, "skill")
    register(TagFactory, "tag")
    register(TeamFactory, "team")
    register(UserFactory, "user")


@pytest.fixture(name="admin_client")
def admin_client_fixture(db, admin_user):
    from worf.testing import ApiClient

    client = ApiClient()
    client.force_login(admin_user)
    return client


@pytest.fixture(name="client")
def client_fixture():
    from worf.testing import ApiClient

    return ApiClient()


@pytest.fixture(name="now")
def now_fixture():
    from django.utils import timezone

    return timezone.now()


@pytest.fixture(name="user_client")
def user_client_fixture(db, user):
    from worf.testing import ApiClient

    client = ApiClient()
    client.force_login(user)
    return client



import marshmallow

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models.fields.files import FieldFile

from worf import fields  # noqa: F401
from worf.casing import snake_to_camel


class SerializerOptions(marshmallow.SchemaOpts):
    def __init__(self, meta, **kwargs):
        super().__init__(meta, **kwargs)

        meta_defaults = getattr(settings, "WORF_SERIALIZER_DEFAULT_OPTIONS", {})

        for key, default_value in meta_defaults.items():
            setattr(self, key, getattr(meta, key, default_value))


class Serializer(marshmallow.Schema):
    """
    Serializers are basically marshmallow schemas with extra handling for
    Django models/managers, some tweaks to field naming/ordering and the option
    to supply some meta defaults via Django settings.

    The read/write methods are relics of our interim serialization strategy
    they're analogous with marshmallow dump/load, but a bit less flexible.

    Marshmallow is primarily used for field filtering right now, worf coerces
    values itself, but we might delegated more to marshmallow later.
    """

    OPTIONS_CLASS = SerializerOptions

    TYPE_MAPPING = {
        **marshmallow.Schema.TYPE_MAPPING,
        FieldFile: fields.File,
    }

    def __repr__(self):
        return f"<{self.__class__.__name__}()>"

    @property
    def dict_class(self):
        return dict

    def list(self, items):
        return [self.read(item) for item in items]

    def on_bind_field(self, field_name, field_obj):
        field_obj.data_key = snake_to_camel(field_obj.data_key or field_name)

    def read(self, obj):
        return self.dump(obj)

    def write(self):
        return list(self.load_fields.keys())

    class Meta:
        ordered = True


class LegacySerializer:
    def __init__(self, model_class, api_method):
        self.api_method = api_method
        self.model_class = model_class

    def __repr__(self):
        return f'<{self.__class__.__name__}(model_class={self.model_class.__name__}, api_method="{self.api_method}")>'

    def list(self, items):
        return [self.read(item) for item in items]

    def read(self, obj):
        payload = getattr(obj, self.api_method)()
        if not isinstance(payload, dict):
            msg = f"{obj.__name__}.{self.api_method}() did not return a dictionary"
            raise ImproperlyConfigured(msg)
        return payload

    def write(self):
        return getattr(self.model_class(), f"{self.api_method}_update_fields")()



import operator
from functools import reduce
import warnings

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Paginator, EmptyPage
from django.db.models import Q

from worf.casing import camel_to_snake
from worf.exceptions import HTTP420
from worf.filters import apply_filterset, generate_filterset
from worf.views.base import AbstractBaseAPI
from worf.views.create import CreateAPI


class ListAPI(AbstractBaseAPI):
    lookup_url_kwarg = "id"  # default incase lookup_field is set
    filters = {}
    ordering = []
    filter_fields = []
    search_fields = []
    sort_fields = []
    queryset = None
    filter_set = None
    list_serializer = None
    count = 0
    page_num = 1
    per_page = 25
    max_per_page = None
    num_pages = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        codepath = self.codepath

        if not isinstance(self.filters, dict):
            raise ImproperlyConfigured(f"{codepath}.filters must be type: dict")

        if not isinstance(self.ordering, list):
            raise ImproperlyConfigured(f"{codepath}.ordering must be type: list")

        if not isinstance(self.filter_fields, list):
            raise ImproperlyConfigured(f"{codepath}.filter_fields must be type: list")

        if not isinstance(self.search_fields, (dict, list)):
            raise ImproperlyConfigured(f"{codepath}.search_fields must be type: list")

        if not isinstance(self.sort_fields, list):
            raise ImproperlyConfigured(f"{codepath}.sort_fields must be type: list")

        # generate a default filterset if a custom one was not provided
        if self.filter_set is None:
            self.filter_set = generate_filterset(self.model, self.queryset)

        # support deprecated search_fields and/or dict syntax (note that `and` does nothing)
        if isinstance(self.search_fields, dict):
            warnings.warn(
                f"Passing a dict to {codepath}.search_fields is deprecated. Pass a list instead."
            )
            self.search_fields = self.search_fields.get("or", [])

    def get(self, request, *args, **kwargs):
        return self.render_to_response()

    def _set_base_lookup_kwargs(self):
        # Filters set directly on the class
        self.lookup_kwargs.update(self.filters)

        # Filters set via URL
        if hasattr(self, "lookup_field") and hasattr(self, "lookup_url_kwarg"):
            self.lookup_kwargs.update(
                {self.lookup_field: self.kwargs[self.lookup_url_kwarg]}
            )

        self.validate_lookup_field_values()

    def set_search_lookup_kwargs(self):
        """
        Set generic lookup kwargs based on q and additional GET params.

        For more advanced search use cases, override this method and pass GET
        with any remaining params you want to use classic django filters for.
        """
        if not self.filter_fields and not self.search_fields:
            return

        self.set_bundle_from_query_string(self.request)

        # Whatever is not q or page as a querystring param will
        # be used for key-value search.
        query = self.bundle.pop("q", "").strip()

        self.bundle.pop("page", None)
        self.bundle.pop("p", None)

        if query:
            search_icontains = (
                Q(**{f"{search_field}__icontains": query})
                for search_field in self.search_fields
            )
            self.search_query = reduce(operator.or_, search_icontains)

        if not self.filter_fields or not self.bundle:
            return

        for key in self.bundle.keys():
            strip_key = key.rstrip("!")

            if strip_key not in self.filter_fields:
                continue

            value = self.bundle[key]

            # support passing in and range as lists
            if isinstance(value, list):
                if strip_key.endswith("__in") or strip_key.endswith("__range"):
                    value = ",".join(str(item) for item in value)

            self.lookup_kwargs[key] = value

    def get_queryset(self):
        if self.queryset is None:
            return self.model.objects.all()
        return self.queryset.all()

    def get_processed_queryset(self):
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  URL Kwargs
        # If these aren't reset they are polluted by cache (somehow)

        self.lookup_kwargs = {}
        self.search_query = Q()

        self._set_base_lookup_kwargs()
        self.set_search_lookup_kwargs()

        order_by = self.ordering
        if self.request.GET.get("sort"):
            sort = self.parse_sort(self.request.GET.getlist("sort"))
            if set([s.lstrip("-") for s in sort]).issubset(self.sort_fields):
                order_by = sort

        lookups = self.lookup_kwargs.items()
        filterset_kwargs = {k: v for k, v in lookups if not isinstance(v, list)}
        list_kwargs = {k: v for k, v in lookups if isinstance(v, list)}

        queryset = self.get_queryset()

        try:
            queryset = (
                apply_filterset(self.filter_set, queryset, filterset_kwargs)
                .filter(self.search_query)
                .distinct()
            )

            for key, value in list_kwargs.items():
                for item in value:
                    queryset = (
                        queryset.exclude(**{key.rstrip("!"): item})
                        if key.endswith("!")
                        else queryset.filter(**{key: item})
                    )

            if order_by:
                queryset = queryset.order_by(*order_by)
        except TypeError as e:
            if settings.DEBUG:
                raise HTTP420(f"Error, {self.lookup_kwargs}, {e.__cause__}")
            raise e

        return queryset

    def get_serializer(self):
        if self.list_serializer and self.request.method == "GET":
            return self.list_serializer()
        return super().get_serializer()

    def paginated_results(self):
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PAGINATION
        queryset = self.get_processed_queryset()
        request = self.request

        if settings.DEBUG:
            self.query = str(queryset.query)

        default_per_page = getattr(self, "results_per_page", self.per_page)
        per_page = max(int(request.GET.get("perPage") or default_per_page), 1)
        max_per_page = self.max_per_page or default_per_page
        paginator = Paginator(queryset, min(per_page, max_per_page))

        self.page_num = int(request.GET.get("page") or request.GET.get("p") or 1)
        if self.page_num < 1:
            self.page_num = 1

        self.num_pages = paginator.num_pages
        self.count = paginator.count

        try:
            return paginator.page(self.page_num)
        except EmptyPage:
            return []

    def specific_fields(self, result):
        fields = self.bundle.get("fields", [])
        if fields:
            return {key: value for key, value in result.items() if key in fields}
        return result

    def parse_sort(self, fields):
        return [self.transform_sort(field) for field in fields]

    def transform_sort(self, field):
        prefix = "-" if field[0] == "-" else ""
        return prefix + "__".join(map(camel_to_snake, field.lstrip("-").split(".")))

    def serialize(self):
        serializer = self.get_serializer()

        payload = {
            str(self.name): [
                self.specific_fields(serializer.read(instance))
                for instance in self.paginated_results()
            ]
        }

        if self.per_page:
            payload.update(
                {
                    "pagination": dict(
                        count=self.count,
                        pages=self.num_pages,
                        page=self.page_num,
                    )
                }
            )

        if not settings.DEBUG:
            return payload

        if not hasattr(self, "lookup_kwargs"):
            # Debug throws an error in the event there are no lookup_kwargs
            self.lookup_kwargs = {}

        payload.update(
            {
                "debug": {
                    "bundle": self.bundle,
                    "lookup_kwargs": self.lookup_kwargs,
                    "query": self.query,
                    "search_query": str(self.search_query),
                    "serializer": str(serializer),
                }
            }
        )

        return payload


class ListCreateAPI(CreateAPI, ListAPI):
    pass



from worf.assigns import AssignAttributes
from worf.views.base import AbstractBaseAPI


class CreateAPI(AssignAttributes, AbstractBaseAPI):
    create_serializer = None

    def create(self):
        instance = self.get_instance()
        self.validate()
        self.save(instance, self.bundle)
        instance.refresh_from_db()
        return instance

    def get_instance(self):
        return self.model()

    def get_serializer(self):
        if self.create_serializer and self.request.method == "POST":
            return self.create_serializer()
        return super().get_serializer()

    def post(self, request, *args, **kwargs):
        new_instance = self.create()
        serializer = self.get_serializer()
        return self.render_to_response(serializer.read(new_instance), 201)



import pytest

from worf.casing import camel_to_snake, snake_to_camel
from worf.exceptions import NamingThingsError


def test_converts_to_camel():
    assert camel_to_snake("camelToSnake") == "camel_to_snake"


def test_converts_to_snake():
    assert snake_to_camel("snake_to_camel") == "snakeToCamel"


def test_camel_to_snake_raises_NamingThingsError():
    with pytest.raises(NamingThingsError):
        camel_to_snake("Not_a_camel")


def test_snake_to_camel_raises_NamingThingsError():
    with pytest.raises(NamingThingsError):
        snake_to_camel("NotASnake")


def test_camel_to_snake_returns_fast_for_lc():
    assert "snake" == camel_to_snake("snake")


def test_camel_to_snake_catches_IDs():
    with pytest.raises(NamingThingsError):
        camel_to_snake("thisIsNotAnValidID")


def test_camel_to_snake_catches_invalid_chars():
    with pytest.raises(NamingThingsError):
        camel_to_snake("thisAintNo🐪")


def test_snake_to_camel_catches_invalid_chars():
    with pytest.raises(NamingThingsError):
        snake_to_camel("this_aint_no_🐍")



from worf.views.base import APIResponse, AbstractBaseAPI  # noqa
from worf.views.create import CreateAPI  # noqa
from worf.views.delete import DeleteAPI  # noqa
from worf.views.detail import DetailAPI, DetailUpdateAPI  # noqa
from worf.views.list import ListAPI, ListCreateAPI  # noqa
from worf.views.update import UpdateAPI  # noqa



from django.db.models import Manager

import marshmallow.fields
from marshmallow.fields import *  # noqa: F401, F403


class File(marshmallow.fields.Field):
    def _serialize(self, value, attr, obj, **kwargs):
        return value.url if value.name else None


class Nested(marshmallow.fields.Nested):
    def _serialize(self, nested_obj, attr, obj, **kwargs):
        if isinstance(nested_obj, Manager):
            nested_obj = nested_obj.all()
        return super()._serialize(nested_obj, attr, obj, **kwargs)



from functools import partialmethod

from django.core.files import File
from django.test.client import Client, MULTIPART_CONTENT

JSON_CONTENT = "application/json"


class ApiClient(Client):
    def generic_multipart(self, method, path, data=None, content_type=None, **kwargs):
        content_type = content_type or self._guess_content_type(data)
        data = self._encode_multipart(data, content_type)
        return self.generic(method, path, data, content_type, **kwargs)

    patch = partialmethod(generic_multipart, "PATCH")
    post = partialmethod(generic_multipart, "POST")
    put = partialmethod(generic_multipart, "PUT")

    def _contains_files(self, data):
        return any(isinstance(value, File) for value in data.values())

    def _encode_multipart(self, data, content_type):
        post_data = self._encode_json({} if data is None else data, content_type)
        return self._encode_data(post_data, content_type)

    def _guess_content_type(self, data):
        return (
            MULTIPART_CONTENT
            if isinstance(data, dict) and self._contains_files(data)
            else JSON_CONTENT
        )



from django.core.exceptions import ValidationError
from django.db import models
from django.db.utils import IntegrityError


class AssignAttributes:
    def save(self, instance, bundle):
        items = [
            (key, getattr(self.model, key), value) for key, value in bundle.items()
        ]

        for key, attr, value in items:
            if isinstance(value, models.Model):
                setattr(instance, key, value)
                continue

            if isinstance(attr.field, models.ForeignKey):
                self.set_foreign_key(instance, key, value)
                continue

            if isinstance(attr.field, models.ManyToManyField):
                continue

            setattr(instance, key, value)

        instance.save()

        for key, attr, value in items:
            if isinstance(attr.field, models.ManyToManyField):
                if not attr.through._meta.auto_created:
                    self.set_many_to_many_with_through(instance, key, value)
                    continue

                self.set_many_to_many(instance, key, value)

    def set_foreign_key(self, instance, key, value):
        related_model = self.get_related_model(key)
        try:
            related_instance = (
                related_model.objects.get(pk=value) if value is not None else None
            )
        except related_model.DoesNotExist as e:
            raise ValidationError(f"Invalid {self.keymap[key]}") from e
        setattr(instance, key, related_instance)

    def set_many_to_many(self, instance, key, value):
        try:
            getattr(instance, key).set(value)
        except (IntegrityError, ValueError) as e:
            raise ValidationError(f"Invalid {self.keymap[key]}") from e

    def set_many_to_many_with_through(self, instance, key, value):
        try:
            attr = getattr(self.model, key)

            through_model = attr.through
            model_name = self.model._meta.model_name
            target_field_name = attr.field.m2m_target_field_name()
            reverse_name = attr.field.m2m_reverse_name()

            getattr(instance, key).clear()

            through_model.objects.bulk_create(
                [
                    through_model(
                        **{
                            item_key: item_value
                            for item_key, item_value in item.items()
                            if item_key != target_field_name
                        },
                        **{
                            model_name: instance,
                            reverse_name: item[target_field_name],
                        },
                    )
                    for item in value
                ]
            )
        except (AttributeError, IntegrityError, ValueError) as e:
            raise ValidationError(f"Invalid {self.keymap[key]}") from e

    def validate(self):
        for key in self.bundle.keys():
            self.validate_bundle(key)

            field = self.model._meta.get_field(key)

            if self.bundle[key] is None and not field.null:
                raise ValidationError(f"Invalid {self.keymap[key]}")



import pytest

from uuid import uuid4

from django.core.exceptions import ValidationError

uuid = uuid4()
email = "something@example.com"
phone = "(555) 555-5555"


@pytest.fixture(name="profile_view")
def profile_view_fixture(db, profile_factory):
    from django.test import RequestFactory

    from tests.views import ProfileDetail

    profile_factory.create(email=email, phone=phone)
    view = ProfileDetail()
    view.bundle = {
        "id": str(uuid),
        "email": email,
        "phone": phone,
    }
    view.request = RequestFactory().patch(f"/{uuid}/")
    view.kwargs = dict(id=str(uuid))
    view.serializer = None
    return view


def test_validate_bundle(profile_view):
    assert profile_view.validate_bundle("id")
    assert profile_view.validate_bundle("email")
    assert profile_view.validate_bundle("phone")


def test_validate_uuid_accepts_str(profile_view):
    assert profile_view.validate_uuid(str(uuid)) == uuid


def test_validate_uuid_accepts_uuid(profile_view):
    assert profile_view.validate_uuid(uuid) == uuid


def test_validate_uuid_raises_error(profile_view):
    with pytest.raises(ValidationError):
        profile_view.validate_uuid("not-a-uuid")


def test_validate_email_passes(profile_view):
    assert profile_view.validate_email(email) == email


def test_validate_email_raises_error(profile_view):
    with pytest.raises(ValidationError):
        profile_view.validate_email("fake.example@com")


def test_validate_custom_field_passes(profile_view):
    assert profile_view.validate_phone(phone) == "+5555555555"


def test_validate_custom_field_raises_error(profile_view):
    with pytest.raises(ValidationError):
        profile_view.validate_phone("invalid number")



from django.urls import include, path

from tests.views import (
    ProfileDetail,
    ProfileList,
    ProfileListSubSet,
    UserDetail,
    UserList,
)

urlpatterns = [
    path("profiles/", ProfileList.as_view()),
    path("profiles/subset/", ProfileListSubSet.as_view()),
    path("profiles/<str:id>/", ProfileDetail.as_view()),
    path("users/", UserList.as_view()),
    path("users/<str:id>/", UserDetail.as_view()),
]



from datetime import datetime
from uuid import UUID

from django.conf import settings
from django.db import models
from django.core.exceptions import ValidationError
from django.core.validators import validate_email
from django.utils.dateparse import parse_datetime

from worf.exceptions import NotImplementedInWorfYet


class ValidationMixin:
    boolean_values = {
        "1": True,
        "0": False,
        "true": True,
        "false": False,
    }

    secure_fields = [
        "password",
        "password_confirmation",
    ]

    def _coerce_bool(self, value):
        if isinstance(value, bool):
            return value

        return self.boolean_values.get(str(value).lower().strip(), None)

    def _validate_boolean(self, key):
        value = self.bundle[key]
        coerced = self._coerce_bool(value)

        if not isinstance(coerced, bool):
            raise ValidationError(
                f"Field {self.keymap[key]} accepts a boolean, got {value}, coerced to {coerced}"
            )

        return coerced

    def _validate_date(self, key):
        value = self.bundle[key]

        if not isinstance(value, str):
            return None

        return datetime.strptime(value, "%Y-%m-%d")

    def _validate_datetime(self, key):
        value = self.bundle[key]
        coerced = None

        if isinstance(value, str):
            coerced = parse_datetime(value)

        if not isinstance(coerced, datetime):
            raise ValidationError(
                f"Field {self.keymap[key]} accepts a iso datetime string, got {value}, coerced to {coerced}"
            )

        return coerced

    def _validate_many_to_many(self, key):
        value = self.bundle[key]

        if not isinstance(value, list):
            raise ValidationError(
                f"Field {self.keymap[key]} accepts an array, got {type(value)} {value}"
            )

    def _validate_string(self, key, max_length):
        value = self.bundle[key]

        if not isinstance(value, str):
            raise ValidationError(f"Field {self.keymap[key]} accepts string")

        if max_length is not None and len(value) > max_length:
            raise ValidationError(
                f"Field {self.keymap[key]} accepts a maximum of {max_length} characters"
            )

        return value

    def _validate_int(self, key):
        value = self.bundle[key]

        if value is None or value == "":
            return None

        try:
            integer = int(value)
        except (TypeError, ValueError):
            raise ValidationError(f"Field {self.keymap[key]} accepts an integer")

        return integer

    def _validate_positive_int(self, key):
        integer = self._validate_int(key)

        if integer < 0:
            raise ValidationError(
                f"Field {self.keymap[key]} accepts a positive integer"
            )

        return integer

    ############################################################################
    # Public methods for use downstream
    def coerce_array_of_integers(self, key):
        try:
            self.bundle[key] = [int(id) for id in self.bundle[key]]
        except ValueError:
            message = f"Field {self.keymap[key]} accepts an array of integers. Got {self.bundle[key]} instead."
            raise ValidationError(message + " I couldn't coerce the values.")

    def validate_int(self, value):
        if not isinstance(value, int):
            raise ValidationError(f"Expected integer, got {value}")

    def validate_numeric(self, value):
        try:
            int(value)
        except ValueError:
            raise ValidationError(f"Expected numeric, got {value}")

    def validate_uuid(self, value):
        if value is None:
            raise ValidationError(f"Expected UUID, got {value}")
        try:
            return UUID(str(value))
        except (TypeError, ValueError):
            raise ValidationError(f"Expected UUID, got {value}")

    def validate_email(self, value):
        if not isinstance(value, str):
            raise ValidationError(f"{value} is not a valid email address")

        email = value.strip().lower()
        try:
            validate_email(email)
        except ValidationError:
            raise ValidationError(f"{value} is not a valid email address")
        return email

    def validate_bundle(self, key):
        """
        Handle basic type validation and coercion.

        @param key: the model attribute to check against.

        @raise NotImplementedInWorfYet: If the field type is not currently supported
        @raise ValidationError: If this is a write and `key` is not serializer editable
        @raise ValidationError: If a value fails to pass validation

        @return: If all checks pass, True is returned.

        Side Effects:
        As various bundle objects are parsed and validated, we reset the bundle.
        This may result in self.bundle changes.

        We expect to set a fully validated bundle keys and values.
        """
        serializer = self.get_serializer()
        write_methods = ("PATCH", "POST", "PUT")

        if self.request.method in write_methods and key not in serializer.write():
            message = f"{self.keymap[key]} is not editable"
            if settings.DEBUG:
                message += f":: {serializer}"
            raise ValidationError(message)

        annotation = (
            self.get_queryset().query.annotations.get(key)
            if hasattr(self, "get_queryset")
            else None
        )

        if not hasattr(self.model, key) and not annotation:
            raise ValidationError(f"{self.keymap[key]} does not exist")

        if key not in self.secure_fields and isinstance(self.bundle[key], str):
            self.bundle[key] = self.bundle[key].strip()

        field = (
            annotation.output_field if annotation else self.model._meta.get_field(key)
        )

        if field.blank and field.empty_strings_allowed and self.bundle[key] == "":
            pass

        elif field.null and self.bundle[key] is None:
            pass

        elif hasattr(self, f"validate_{key}"):
            self.bundle[key] = getattr(self, f"validate_{key}")(self.bundle[key])

        elif isinstance(field, models.UUIDField):
            self.bundle[key] = self.validate_uuid(self.bundle[key])

        elif isinstance(field, (models.CharField, models.TextField, models.SlugField)):
            self.bundle[key] = self._validate_string(key, field.max_length)

        elif isinstance(field, models.EmailField):
            self.bundle[key] = self.validate_email(self.bundle[key])

        elif isinstance(field, (models.IntegerField, models.SmallIntegerField)):
            # TODO check size of SmallIntegerField
            self.bundle[key] = self._validate_int(key)

        elif isinstance(field, models.PositiveIntegerField):
            self.bundle[key] = self._validate_positive_int(key)

        elif isinstance(field, models.ManyToManyField):
            self._validate_many_to_many(key)

        elif isinstance(field, models.BooleanField):
            self.bundle[key] = self._validate_boolean(key)

        elif isinstance(field, models.FileField):
            pass  # Django will raise an exception if handled improperly

        elif isinstance(field, models.ForeignKey):
            pass  # Django will raise an exception if handled improperly

        elif isinstance(field, models.DateTimeField):
            self.bundle[key] = self._validate_datetime(key)

        elif isinstance(field, models.DateField):
            self.bundle[key] = self._validate_date(key)

        elif isinstance(field, models.JSONField):
            pass
            # TODO check the type of json object we're expecting
            # try:
            #     json.loads(self.bundle[key])
            # except ValueError:
            #     raise ValidationError(f"Field {self.keymap[key]} requires valid JSON")

        else:
            message = f"{field.get_internal_type()} has no validation method for {key}"
            if settings.DEBUG:
                message += f":: Received {self.bundle[key]}"
            raise NotImplementedInWorfYet(message)
            # TODO
            # FileField
            # UUIDField
            # We don't handle 1:1 or FKs.

        return True



import os
from setuptools import setup, find_packages


def read(rel_path):
    # type: (str) -> str
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path):
    # type: (str) -> str
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


long_description = read("README.md")

setup(
    author_email="wade@wadewilliams.com",
    author="Wade Williams",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    description="Worf (Wade's Own Rest Framework): A more Djangonic approach to REST APIs",
    install_requires=[
        "Django>=3.0.0,<3.3",
        "django-url-filter>=0.3.15",
        "marshmallow>=3.14.0",
    ],
    include_package_data=True,
    keywords="django, rest, framework, api",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="worf",
    packages=find_packages(),
    python_requires=">=3.7",
    url="https://github.com/gundotio/worf",
    version=get_version("worf/__init__.py"),
    zip_safe=False,
)



from django.core.exceptions import ValidationError
from django.contrib.auth.models import User
from django.db.models import F, Value
from django.db.models.functions import Concat

from worf.permissions import PublicEndpoint
from worf.views import CreateAPI, DeleteAPI, DetailAPI, ListAPI, UpdateAPI

from tests.models import Profile
from tests.serializers import ProfileSerializer, UserSerializer


class ProfileList(CreateAPI, ListAPI):
    model = Profile
    queryset = Profile.objects.annotate(
        name=Concat("user__first_name", Value(" "), "user__last_name"),
        date_joined=F("user__date_joined"),
    )
    ordering = ["pk"]
    serializer = ProfileSerializer
    permissions = [PublicEndpoint]
    search_fields = []
    filter_fields = [
        "name",
        "name__icontains",
        "name__in",
        "date_joined__gte",
        "tags",
        "tags__in",
    ]


class ProfileListSubSet(ProfileList):
    queryset = ProfileList.queryset.none()


class ProfileDetail(DeleteAPI, UpdateAPI, DetailAPI):
    model = Profile
    serializer = ProfileSerializer
    permissions = [PublicEndpoint]

    def validate_phone(self, value):
        try:
            assert value == "(555) 555-5555"
        except AssertionError:
            raise ValidationError("{value} is not a valid phone number")
        return "+5555555555"


class UserList(ListAPI):
    model = User
    ordering = ["pk"]
    serializer = UserSerializer
    permissions = [PublicEndpoint]
    filter_fields = [
        "email",
        "date_joined__gte",
        "date_joined__lte",
    ]
    search_fields = [
        "email",
        "username",
    ]
    sort_fields = [
        "id",
        "date_joined",
    ]


class UserDetail(UpdateAPI, DetailAPI):
    model = User
    serializer = UserSerializer
    permissions = [PublicEndpoint]



from django.core.exceptions import ImproperlyConfigured

from worf.lookups import FindInstance
from worf.views.base import AbstractBaseAPI
from worf.views.update import UpdateAPI


class DetailAPI(FindInstance, AbstractBaseAPI):
    detail_serializer = None

    def get(self, request, *args, **kwargs):
        return self.render_to_response()

    def get_serializer(self):
        if self.detail_serializer and self.request.method == "GET":
            return self.detail_serializer()
        return super().get_serializer()

    def serialize(self):
        """Return the model api, used for responses."""
        serializer = self.get_serializer()
        payload = serializer.read(self.get_instance())
        if not isinstance(payload, dict):
            raise ImproperlyConfigured(f"{serializer} did not return a dictionary")
        return payload


class DetailUpdateAPI(UpdateAPI, DetailAPI):
    def patch(self, request, *args, **kwargs):
        self.update()
        return self.get(request)



import json
import types
import warnings

from io import BytesIO
from urllib.parse import parse_qs

from django.conf import settings
from django.core.exceptions import (
    ImproperlyConfigured,
    ObjectDoesNotExist,
    ValidationError,
)
from django.db import models
from django.http import HttpResponse, JsonResponse
from django.template.response import TemplateResponse
from django.views import View
from django.views.decorators.cache import never_cache
from django.utils.decorators import method_decorator

from worf.casing import camel_to_snake, snake_to_camel
from worf.exceptions import HTTP_EXCEPTIONS, HTTP404, HTTP422, PermissionsException
from worf.serializers import LegacySerializer
from worf.validators import ValidationMixin

api_name = getattr(settings, "WORF_API_NAME", "Worf API")
api_root = getattr(settings, "WORF_API_ROOT", "/api/")


@method_decorator(never_cache, name="dispatch")
class APIResponse(View):
    def __init__(self, *args, **kwargs):
        self.codepath = f"{self.__module__}.{self.__class__.__name__}"
        return super().__init__(*args, **kwargs)

    def serialize(self):
        raise NotImplementedError

    def render_to_response(self, data=None, status_code=200):
        if data is None:
            data = self.serialize()

        if data is None:
            msg = f"{self.codepath} did not pass an object to "
            msg += "render_to_response, nor did its serializer method"
            raise ImproperlyConfigured(msg)

        is_html_request = (
            "text/html" in self.request.headers.get("Accept", "")
            and self.request.GET.get("format") != "json"
        )

        json_kwargs = dict(json_dumps_params=dict(indent=2)) if is_html_request else {}

        response = JsonResponse(data, **json_kwargs) if data != "" else HttpResponse()
        response.status_code = status_code

        if is_html_request:
            template = "worf/api.html"
            context = dict(
                api_name=api_name,
                api_root=api_root,
                content=response.content.decode("utf-8"),
                response=response,
            )
            response = TemplateResponse(self.request, template, context=context)
            response.status_code = status_code
            response.render()

        return response


class AbstractBaseAPI(APIResponse, ValidationMixin):
    model = None
    permissions = []
    api_method = "api"
    serializer = None
    payload_key = None

    def __init__(self, *args, **kwargs):
        self.codepath = f"{self.__module__}.{self.__class__.__name__}"

        if self.model is None:
            raise ImproperlyConfigured(f"Model is not set on {self.codepath}")

        if not isinstance(self.permissions, list):
            raise ImproperlyConfigured(
                f"{self.codepath}.permissions must be type: list"
            )

        for method in ["post", "patch", "put", "delete"]:
            if method in dir(self) and not len(self.permissions):
                warnings.warn(
                    "\n{} method allowed on {} without permissions.".format(
                        method.upper(),
                        self.codepath,
                    ),
                )

        for method in self.permissions:
            # append authorization functions to this class
            setattr(self, method.__name__, types.MethodType(method, self))

        super().__init__(*args, **kwargs)

    @property
    def name(self):
        if isinstance(self.payload_key, str):
            return self.payload_key
        verbose_name_plural = self.model._meta.verbose_name_plural
        return snake_to_camel(verbose_name_plural.replace(" ", "_").lower())

    def _check_permissions(self):
        """Return a permissions exception when in debug mode instead of 404."""
        for method in self.permissions:
            permission_func = getattr(self, method.__name__)
            response = permission_func(self.request)
            if response == 200:
                continue

            if settings.DEBUG:
                raise PermissionsException(
                    "Permissions function {}.{} returned {}. You'd normally see a 404 here but DEBUG=True.".format(
                        method.__module__, method.__name__, response
                    )
                )

            try:
                raise response
            except TypeError:
                raise ImproperlyConfigured(
                    "Permissions function {}.{} must return 200 or an HTTPException".format(
                        method.__module__, method.__name__
                    )
                )

    def _get_lookup_field(self, field):
        related = field.find("__")

        """Support one level of related field reference."""
        if related != -1:
            related_field = field[:related]
            target_field = field[related + 2 :]

            if target_field in ["gt", "lt", "contains", "startswith", "gte", "lte"]:
                return False

            if target_field.find("__") != -1:
                return False

            return self.get_related_model(related_field)._meta.get_field(target_field)
            # TODO if there is another reference, recurse

        return self.model._meta.get_field(field)

    def get_related_model(self, field):
        return self.model._meta.get_field(field).related_model

    def get_serializer(self):
        if self.serializer:
            return self.serializer()
        if self.api_method:
            return LegacySerializer(self.model, self.api_method)
        msg = f"{type(self).__name__}.get_serializer() did not return a serializer"
        raise ImproperlyConfigured(msg)

    def validate_lookup_field_values(self):
        # todo check for each lookup kwarg
        for field, url_kwarg in self.lookup_kwargs.items():
            lookup_field = self._get_lookup_field(field)

            if isinstance(lookup_field, models.UUIDField):
                self.validate_uuid(url_kwarg)
            elif isinstance(
                lookup_field,
                (
                    models.ForeignKey,
                    models.IntegerField,
                    models.PositiveIntegerField,
                    models.SmallIntegerField,
                ),
            ):
                self.validate_numeric(url_kwarg)

    def flatten_bundle(self, raw_bundle):
        # parse_qs gives us a dictionary where all values are lists
        return {
            key: value[0] if len(value) == 1 else value
            for key, value in raw_bundle.items()
        }

    def get_parts(self, request):
        if request.method == "POST":
            return request.POST, request.FILES

        return request.parse_file_upload(request.META, BytesIO(request.body))

    def set_bundle(self, raw_bundle):
        self.bundle = {}
        self.keymap = {}

        if not raw_bundle:
            return

        for key in raw_bundle.keys():
            field = camel_to_snake(key)
            self.bundle[field] = raw_bundle[key]
            self.keymap[field] = key

    def set_bundle_from_query_string(self, request):
        raw_bundle = self.flatten_bundle(parse_qs(request.META["QUERY_STRING"]))

        self.set_bundle(raw_bundle)

    def set_bundle_from_request_body(self, request):
        raw_bundle = {}

        if request.content_type == "multipart/form-data":
            post, files = self.get_parts(request)
            raw_bundle.update(self.flatten_bundle(post))
            raw_bundle.update(self.flatten_bundle(files))
        elif request.body:
            try:
                raw_bundle = json.loads(request.body)
            except json.decoder.JSONDecodeError:
                pass

        self.set_bundle(raw_bundle)

    def dispatch(self, request, *args, **kwargs):
        method = request.method.lower()
        handler = self.http_method_not_allowed

        if method in self.http_method_names:
            handler = getattr(self, method, self.http_method_not_allowed)

        try:
            self._check_permissions()  # only returns 200 or HTTP_EXCEPTIONS
            self.set_bundle_from_request_body(request)
            return handler(request, *args, **kwargs)  # calls self.serialize()
        except HTTP_EXCEPTIONS as e:
            return self.render_to_response(dict(message=e.message), e.status)
        except ObjectDoesNotExist:
            return self.render_to_response(
                dict(message=HTTP404.message), HTTP404.status
            )
        except ValidationError as e:
            return self.render_to_response(dict(message=e.message), HTTP422.status)



from worf.views.base import AbstractBaseAPI


class DeleteAPI(AbstractBaseAPI):
    def delete(self, request, *args, **kwargs):
        self.destroy()
        return self.render_to_response("", 204)

    def destroy(self):
        self.get_instance().delete()



from uuid import uuid4

from django.db import models
from django.contrib.auth.models import User


class Profile(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid4)

    avatar = models.FileField(upload_to="avatars/", blank=True)
    email = models.CharField(max_length=64)
    phone = models.CharField(max_length=64)

    user = models.OneToOneField(User, on_delete=models.CASCADE)
    role = models.ForeignKey("Role", on_delete=models.CASCADE)
    team = models.ForeignKey("Team", blank=True, null=True, on_delete=models.SET_NULL)
    skills = models.ManyToManyField("Skill", through="RatedSkill")
    tags = models.ManyToManyField("Tag")

    def api(self):
        return dict(id=self.id, email=self.email, phone=self.phone)

    def api_update_fields(self):
        return [
            "id",
            "email",
            "phone",
        ]


class Role(models.Model):
    name = models.CharField(max_length=64)


class Skill(models.Model):
    name = models.CharField(max_length=200)


class RatedSkill(models.Model):
    class Meta:
        ordering = ["skill__name"]
        unique_together = ["profile", "skill"]

    RATING_CHOICES = [
        (5, "Excellent"),
        (4, "Great"),
        (3, "Average"),
        (2, "Poor"),
        (1, "Bad"),
    ]

    profile = models.ForeignKey("Profile", on_delete=models.CASCADE)
    rating = models.IntegerField(choices=RATING_CHOICES)
    skill = models.ForeignKey("Skill", on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.skill.name}: {self.get_rating_display()}"


class Tag(models.Model):
    name = models.CharField(max_length=200)


class Team(models.Model):
    name = models.CharField(max_length=64)
